# -*- coding: utf-8 -*-

"""
nat_mch_telnet.py
~~~~~~~~~~~~~~~~~

This is a lazy implementation of the GenDev interface. The ultimate purpose
of this module is to be encapsulated within another module that really
implements that interface.
Unfortunately, NAT MCHs don't allow to go for a full implementation of the
GenDev interface based on an only communication interface.
The main operation that is only supported by the command line interface is the
firmware update. So that, this module based on Telnet, as access way, and the
command line interface of the MCH, shall only be used to perform a firmware
update and nothing else. For other operations, better use the module based on
the communication via the MHC web interface.
"""

from __future__ import annotations
import re
import time
import socket
import ipaddress
from collections import OrderedDict
from ..gendev_err import (
    ConnTimeout,
    NoRouteToDevice,
    FeatureNotSupported,
    ConnectionRefused,
)
from ..gendev_logger import GenDevLogger
from telnetlib import Telnet

__author__ = ["Felipe Torres González", "Ross Elliot"]
__copyright__ = "Copyright 2021, ESS MTCA Tools"
__credits__ = ["Jeong Han Lee"]
__license__ = "GPL-3.0"
__version__ = "1.1"
__maintainer__ = "Ross ELliot"
__email__ = "ross.elliot@ess.eu"
__status__ = "Development"


class NATMCHTelnet:
    """NATMCTelnet access an NAT MCH via Telnet.

    This module implements some operations using the command line interface
    via Telnet. Thus, the MCH has to be accessible in the network.
    The firmware update relies on **the mchconfig-server** to serve the
    firmware image using the FTP protocol.

    Supported operations:

    - Retrieve the general information of the MCH.
    - Firmware update of the MCH.
    - Enable DHCP mode of the MCH.
    """

    # Telnet prompt
    PROMPT = b"nat> "
    BOOT_TIME = 60

    def __init__(
        self,
        ip_address: str,
        backplane,
        port: int = 23,
        logger: GenDevLogger = None,
        hostname: str = None,
        open: bool = True,
        **kwargs,
    ):
        """Class constructor.

        Args:
            ip_address(str): the IP address of the MCH.
            backplane(BackplaneType): type of backplane in which the MCH is installed.
            port(int): port of the Telnet service (usually, 23).
            logger(Logger): reference to a logger that is being used.
            log_to_file(Bool): boolean flag to indicate if the log should be written
            to a file.
            hostname(str): hostname of the MCH.
            open(bool): whether to establish the connection at instantiation time or later. \
            Set this argument to False when a deferred link establishment is aimed.
            **telnet_ip_address(str): the ip address of the telnet \
            connection, if not equal to the IP address of the MCH, i.e. when \
            connecting via a MOXA (optional).
            **tftp_server_address(str): IP address of the machine hosting the TFTP \
            server with the fw images. Default value = **172.30.4.69**.
            **tftp_server_fw_path(str): path to the fw binaries in the TFTP server. \
            Default value = "fw/".

        Raises:
            ConnTimeout: if the connection to the device gets lost.
            NoRouteToDevice: if the device is not reachable.
        """
        self.ip_address = ip_address
        self.backplane = backplane
        self.hostname = hostname
        self.port = port

        # Optional arguments

        # If no telnet IP address is provided, use MCH IP address
        if "telnet_ip_address" in kwargs:
            self._telnet_ip_address = kwargs["telnet_ip_address"]
        else:
            self._telnet_ip_address = self.ip_address

        if "tftp_server_address" in kwargs:
            self._server_ip = kwargs["tftp_server_address"]
        else:
            self._server_ip = "172.30.4.69"

        if "tftp_server_fw_path" in kwargs:
            self._fw_path = kwargs["tftp_server_fw_path"]
        else:
            self._fw_path = "fw/"

        if "log_to_file" in kwargs:
            self._log_to_file = kwargs["log_to_file"]
        else:
            self._log_to_file = False

        # Initialise Logger instance
        if isinstance(logger, (GenDevLogger)):
            self.logger = logger
        else:
            self.logger = GenDevLogger(logger=logger, log_to_file=self._log_to_file)

        self.logger.logger.info(
            "NATMCHTelnet - NAT MCH Telnet instance created."
            "\tMCH IP Address: {}".format(self.ip_address)
        )

        if open:
            self.open()
            self._session_open = True
        else:
            self._session = Telnet()
            self._session_open = False

        # Regular expresions for extracting the infomration relative to the
        # MCH from the version command.
        self._match_fw_ver = re.compile(r"Firmware (V\d{1,2}\.\d{1,2}\.\d{1,2})")
        # Search for the first occurrence of the token FPGA
        self._match_fpga_ver = re.compile(r"FPGA (V\d{1,2}\.\d{1,2})")
        self._match_mcu_ver = re.compile(r"AVR (\d{1,2}\.\d{1,2})")
        self._match_board_sn = re.compile(r"sn: (\d{6}-\d{4})")
        self._match_ip_addr = re.compile(r"ip address +: +((\d{1,3}\.?){4})")
        self._match_mac_addr = re.compile(r"ieee address +: +(([\d\D]{2}:?){6})")
        self._match_subnet_mask = re.compile(r"network mask +: +((\d{1,3}\.?){4})")
        self._match_gateway_addr = re.compile(r"default gateway +: +((\d{1,3}\.?){4})")
        self._match_hostname = re.compile("hostname +: (.+?(?=\r))")
        self._match_dhcp_state = re.compile("dhcp state +: (.+?(?=\r))")

    def open(self):
        """Establish the connection to the target MCH.

        Raises:
            ConnTimeout: if the connection to the device gets lost.
            NoRouteToDevice: if the device is not reachable.
            ConnectionRefused: if the telnet connection is refused.
        """
        self.logger.logger.debug(
            "Opening Telnet connection to {}:{}.".format(
                self._telnet_ip_address, self.port
            )
        )
        try:
            self._session = Telnet(self._telnet_ip_address, self.port, timeout=10)
            self._session_open = True
        except Exception as e:
            if isinstance(e, socket.timeout):
                msg = "Timeout while opening the link to the MCH using Telnet."
                self.logger.logger.error(msg)
                raise ConnTimeout(msg)
            elif isinstance(e, OSError) and (e.errno == 113 or e.errno == 101):
                msg = "Check the connectivity to the MCH" " using the IP: {}.".format(
                    self._telnet_ip_address
                )
                self.logger.logger.error(msg)
                raise NoRouteToDevice(msg)
            elif isinstance(e, OSError) and (e.errno == 111):
                msg = (
                    "Cannot connect to Telnet port {}:{} "
                    "Please close any existing "
                    "connections.".format(self._telnet_ip_address, self.port)
                )
                self.logger.logger.error(msg)
                raise ConnectionRefused(msg)
            else:
                self.logger.logger.error("Unhandled exception: {}".format(e.strerror))
                raise e

    def close(self):
        """Close the Telnet session.

        This is needed when multiple methods of this module are called in a row.
        """
        self.logger.logger.debug(
            "Closing Telnet connection to {}:{}.".format(self.ip_address, self.port)
        )
        self._session.close()
        self._session_open = False

    def _send_backspace(self, sleep: float = 0.25):
        """Internal method to write a backspace character to the MCH console.

        Some fields in the MCH interface are already populated, and must be
        cleared before writing the new value. The only way to do this is to
        issue a backspace character.
        """
        # Backspace is ASCII character 0x08, or special character ''\b'
        self._session.write(b"\b")
        time.sleep(sleep)

    def _send_command(self, command: str, sleep: int = 1, clear_buffer: bool = True):
        """Internal method for sending a low level command to the MCH.

        This command allows forgetting about the particular details of using
        a Telnet session behind the scenes. A regular command from the MCH
        command line interface can be sent through this interface without
        worrying about the underlying communication.

        Args:
            command: command to be sent to the MCH.
            sleep: amount of seconds to wait after sending a command.
            clear_buffer: send a carriage return before the command. This
            helps clearing previous garbage from the buffer, but
            it should be used with caution because there are
            commands that doesn't expect a carriage return after.
        """
        # clean up
        if clear_buffer:
            self._session.write(b"\r")
            self._session.read_until(self.PROMPT)
            time.sleep(sleep)
        self._session.write(command.encode("ascii") + b"\r")
        time.sleep(sleep)

    def _reboot(self, sleep: int = 70):
        """Internal command to send a reboot to the MCH.

        Args:
            sleep: indicates how many seconds to wait after returning from the
            method. Write a 0 to avoid it.
        """
        self.logger.logger.info("Rebooting MCH device.")
        if self._session_open is False:
            self.open()
        self._send_command("reboot")
        self.close()
        # self._session_open = False
        time.sleep(sleep)

    def _read_command(self) -> str:
        """Internal command to read the Telnet Rx buffer.

        This method attempts to read the content from the buffer without I/O
        blocking.

        Returns:
            A string containing the content of the Rx buffer.
        """
        response = self._session.read_very_eager()
        return response.decode("ascii")

    def _set_hostname(self):
        """Sets the hostname of the MCH device.

        Returns:
            True on success.
            False on failure.
        """
        self.logger.logger.info("Setting hostname to {}".format(self.hostname))

        if self._session_open is False:
            self.open()

        self._send_command("mchcfg")
        # Modify
        self._send_command("11", clear_buffer=False)
        # Clear previous entry
        for _ in range(0, 75):
            self._send_backspace()
        self._send_command(self.hostname, clear_buffer=False)
        self._send_command("q", clear_buffer=False)

        # TODO: re-enable this check
        # success = self._check_hostname()
        success = True

        self.close()

        return success

    def _check_hostname(self) -> bool:
        """Checks that the hostname set on the MCH matches
        the value provided.

        Returns:
            True if the hostname setting matches the expected string.
            False if the hostname setting does not match the expected string.
        """

        if self._session_open is False:
            self.open()

        self._send_command("ni")
        network_info = self._read_command()

        mch_hostname = self._match_hostname.search(network_info).group(1)

        success = False
        if mch_hostname == self.hostname:
            success = True

        self.close()

        return success

    def _enable_dhcp(self):
        """Enables DHCP mode on the MCH

        Returns:
            True if DHCP was enabled successfully.
            False if DHCP setting failed.
        """

        self._send_command("mchcfg")
        self._send_command("3", clear_buffer=False)
        for _ in range(0, 5):
            self._send_command("", clear_buffer=False)
        self._send_command("2", clear_buffer=False)
        for _ in range(0, 7):
            self._send_command("", clear_buffer=False)
        self._send_command("q", clear_buffer=False)

        # The MCH needs a reboot for the hostname
        # to update.
        self._reboot()
        # Wait for the MCH to complete the reboot process
        time.sleep(self.BOOT_TIME)

        # Finally, verify the change
        success = self._check_dhcp()

        return success

    def _check_dhcp(self):
        """Check wether DHCP is enabled.

        Returns:
            True if DHCP is enabled
            False if DHCP is disabled
        """

        if self._session_open is False:
            self.open()

        self._send_command("ni")
        network_info = self._read_command()

        dhcp_state = self._match_dhcp_state.search(network_info).group(1)
        enabled = False
        if dhcp_state == " enabled":
            enabled = True

        self.close()

        return enabled

    def _clear_ip_address(self):
        """Helper method to send enough backspace characters
        to clear an IP address
        """

        # Maximum length of an IP address is 15 characters
        for _ in range(0, 15):
            self._send_backspace()

    def _set_ip_addr(self):
        """Set the IP address of the MCH device."""

        if self._session_open is False:
            self.open()

        # Get network information
        self._send_command("ni")
        network_info = self._read_command()

        # Retrieve gateway address and netmask from network info
        gateway_addr = self._match_gateway_addr.search(network_info).group(1)
        netmask = self._match_subnet_mask.search(network_info).group(1)

        # Calculate broadcast ip_address
        ipv4net = ipaddress.IPv4Network(self.ip_address + "/" + netmask, False)
        broadcast_addr = ipv4net.broadcast_address.compressed

        # Trigger IP setting menu
        self._send_command("ip")

        # Clear the existing IP address, and update with new value
        self._clear_ip_address()
        self._send_command(self.ip_address, clear_buffer=False)
        # Clear the existing netmask, and update with new value
        self._clear_ip_address()
        self._send_command(netmask, clear_buffer=False)
        # Clear the existing broadcast address, and update with new value
        self._clear_ip_address()
        self._send_command(broadcast_addr, clear_buffer=False)
        # Clear the existing gateway address, and update with new value
        self._clear_ip_address()
        self._send_command(gateway_addr, clear_buffer=False)

        # Check we are at the confirmation prompt
        response = self._read_command()
        if not response.endswith("Are you really sure ?"):
            self.close()
            return False  # Failure

        # Issue confirmation 'y'
        self._send_command("y", clear_buffer=False)

        self.close()

        # Success
        return True

    def device_info(self) -> tuple[bool, OrderedDict]:
        """Retrieve the main information about the device.

        The information is returned in a dictionary with 2 categories:
        *Board* and *Network*.
        This feature is supported by all the implemented communication
        interfaces, so the best one is chosen when multiple are allowed.

        An example of the returned dictionary::

            {
              'Board': {
                'fw_ver': 'V2.21.8',
                'fpga_ver': 'V1.14',
                'mcu_ver': '1.2',
                'serial_num': '113522-1426'},
              'Network': {'ip_address': '172.30.5.238',
                'mac_address': '00:40:42:22:05:92',
                'subnet_address': '255.255.252.0',
                'gateway_address': '172.30.7.254'}
            }

        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, a dictionary with the device information.
            - On failure, an empty dictionary.

        Raises:
            ConnTimeout: if the connection to the device gets lost.
            NoRouteToDevice: if the device is not reachable.
        """
        if self._session_open is False:
            self.open()

        self._send_command("version")
        raw_info_version = self._read_command()
        self._send_command("ni")
        raw_info_network = self._read_command()

        if raw_info_version != "" and raw_info_network != "":
            resp_dict = OrderedDict()  # type: OrderedDict
            resp_dict["Board"] = dict()

            resp_dict["Board"]["fw_ver"] = self._match_fw_ver.search(
                raw_info_version
            ).group(1)
            resp_dict["Board"]["fpga_ver"] = self._match_fpga_ver.search(
                raw_info_version
            ).group(1)
            resp_dict["Board"]["mcu_ver"] = self._match_mcu_ver.search(
                raw_info_version
            ).group(1)
            resp_dict["Board"]["serial_num"] = self._match_board_sn.search(
                raw_info_version
            ).group(1)

            resp_dict["Network"] = dict()
            resp_dict["Network"]["ip_address"] = self._match_ip_addr.search(
                raw_info_network
            ).group(1)
            resp_dict["Network"]["mac_address"] = self._match_mac_addr.search(
                raw_info_network
            ).group(1)
            resp_dict["Network"]["subnet_address"] = self._match_subnet_mask.search(
                raw_info_network
            ).group(1)
            resp_dict["Network"]["gateway_address"] = self._match_gateway_addr.search(
                raw_info_network
            ).group(1)
            valid = True
        else:
            valid = False
            resp_dict = OrderedDict()

        self.close()

        return valid, resp_dict

    def set_dhcp_mode(self) -> tuple[bool, str]:
        """Enables DHCP mode in the network configuration of the device.

        Performs the following steps:
            - Enables DHCP mode on the MCH.
            - Sets the internal IP address value to match the address
              provided by the DHCP server(required to prevent DHCP lease
              issues).
            - Sets the hostname value.

        Returns
            If failure, a tuple containing False, and a message about the
            failure.
            If success, a tuple containing True, and an empty string.

        Raises:
            ConnectionError: If the device is not accessible.
            NoValidConn: If no valid connection types supporting this feature
                         are used by the device.
        """
        self.logger.logger.info("Enabling DHCP mode.")

        if self._session_open is False:
            self.open()

        total_success = True
        success = [True] * 3
        response = ""

        # Enable DHCP mode
        success[0] = self._enable_dhcp()
        if not success[0]:
            response = "Enabling of DHCP mode failed.\r\n"

        # Set the internal IP address
        success[1] = self._set_ip_addr()
        if not success[1]:
            response = response + "Setting IP address failed.\r\n"

        # Set the hostname
        success[2] = self._set_hostname()
        if not success[2]:
            response = response + "Setting hostname failed.\r\n"

        # If any of the subtasks failed, fail overall
        if False in success:
            total_success = False

        self.close()

        return total_success, response

    def update_fw(self, fw_version: str, part: str = "MCH") -> tuple[bool, str]:
        """Update the firmware of the device.

        This method expects the firmware binary pointed by the value of the
        argument *fw_version* to be available in the TFTP server.
        Mainly, this method injects the command *update_firmware* to an NAT
        MCH.

        Args:
            fw_version: version release number for the new fw.
            part: not used

        Returns:
            If failure, it returns a tuple containing False, and a message
            about the failure.
            If success, it returns True,
        """
        self.logger.logger.info("Updating MCH firmware")

        if self._session_open is False:
            self.open()

        self._send_command("update_firmware")
        # Avoid clearing the buffer bewteen these commands because it would
        # skip the update mode in the MCH.
        self._send_command(
            "{}:{}{}/mch_fw_{}.bin".format(
                self._server_ip, self._fw_path, fw_version, fw_version
            ),
            clear_buffer=False,
        )
        # Erasing the internal memory. If it is attempted to read now from the
        # buffer, it will get the promt.
        time.sleep(30)
        # There's a useless promt which is received first, get rid of it, and
        # wait for the good one that should come when the flashing is finished.
        response_b = self._session.read_until(b"nat> ")  # type: bytes
        # Sometimes, at this point, the buffer has content, sometimes not.
        # It seems reasonable using a length 100 to detect this situation.
        if len(response_b) < 100:
            response_b = self._session.read_until(b"nat> ")
        response = response_b.decode("ascii")  # type: str

        # Let's see if the update was successful. The MCH prints the word
        # "successful" at the end of the process, just before the prompt.
        if "successful" in response:
            success = (True, "")
            self._reboot()
            # Finally, wait for the MCH to complete the reboot process
            time.sleep(50)
        else:
            # Something went wrong, let's check it!
            if "TFTP: could not get file" in response:
                # This error is mainly caused when the target fw_version
                # is not available in the TFTP server.
                success = (
                    False,
                    "The fw version {} couldn't be found in the"
                    " TFTP server".format(fw_version),
                )
            else:
                success = False, "Unknown error. Check the debug log."

        self.close()

        return success

    def set_configuration(
        self, category: str, data: OrderedDict, apply: bool = True
    ) -> tuple[bool, str]:
        """Change the configuration of the device.

        This method focuses on the configuration parameters that are not
        defined within a configuration script. Specifying the entire set of
        parameters is not mandatory, and also, a particular category of
        settings can be modified without affecting the rest.

        This method supports the following configuration categories (taken
        from the webpage names):

        - Backplane configuration [backplane]

        The configuration files are hosted by the server, and they are named
        following this way: `latest_mch_conf_<form factor>_<modifier>_cfg.txt`.
        For factor should match the type of chassis, while the modifier allows
        choosing a non-standard configuration.

        Args:
            category(str): the target set of parameters to be affected by the change. \
            The accepted values are previously listed.
            data(dic): dictionary containing modifiers that allow for choosing a particular \
            backplane configuration file.
            apply(bool): whether to reboot the MCH to make active the changes or not. \
            Disable rebooting when multiple changes are going to be performed in the device. \
            Reboot  the device after the last change.

        Returns:
            A tuple containing a `bool` indicating if the operation was successful,
            and an error message.

        Raises:
            ConnectionError: If the device is not accessible.
            NoRouteToDevice: when the device is not reachable after applying the \
            new configuration.
        """
        if category != "backplane":
            raise FeatureNotSupported(
                "The Telnet backend only supports changing "
                "the backplane configuration."
            )
        if not isinstance(data, dict):
            return (False, "Wrong data type passed.")

        if self._session_open is False:
            self.open()

        option = "generic"
        if "Backplane configuration" in data:
            if "option" in data["Backplane configuration"]:
                option = data["Backplane configuration"]["option"]

        self._send_command("upload_cfg", clear_buffer=True)
        # Provide enough timeout in case the network is slow
        self._send_command(
            "{}:latest_mch_conf_{}_{}_cfg.txt".format(
                self._server_ip,
                "{}u".format(int(self.backplane)),
                option,
            ),
            sleep=3,
            clear_buffer=False,
        )
        # Read the buffer to check whether the file was found in the TFTP server
        validfile = self._read_command()
        if "TFTP: getting file done" not in validfile:
            self.close()
            return False, "The backplane file wasn't found in the server. See the log."

        self._send_command("y", clear_buffer=False)
        self._send_command("", clear_buffer=True)

        if apply:
            self._reboot()

        self.close()

        return (
            True,
            "Expect some delay until the MCH is accessible in the network again.",
        )

    def _get_slot_state(self, amc_num, close_conn: bool = True):
        """Internal method to get the current state of an
        AMC slot.

        Args:
            amc_num(int): AMC slot number.
            close_conn(bool): Whether to close the telnet
               session when returning, or not.
               Default: True

        Returns:
            A string containing the state code {M1, M2, ... M7}
        """
        if self._session_open is False:
            self.open()

        # Get fru_info for all AMCs
        self._send_command("show_fru")
        fru_info = self._read_command()

        # Parse fru_info for line containing the requested AMC, and parse
        # the state
        match = re.search(r"AMC{}(\s+)(M\d)".format(amc_num), fru_info)

        # Check if the Telnet session should be closed
        if close_conn is True:
            self.close()

        if match is None:
            success = False
            state = "Unable to determine state of AMC {}. " "Check AMC number.".format(
                amc_num
            )
        else:
            success = True
            state = match.group(2)

        # State is in group 2 of returned regex search
        return (success, state)

    def _start_slot(self, amc_num, sleep: int = 2, upstream=False):
        """Internal method to start the card in an AMC slot

        Args:
            amc_num(int): AMC slot number to power on.
            sleep(int): Time to allow for the AMC to power down.
                        Default: 2 seconds

        Returns:
            True if successful
            False if unsuccessful
        """
        success = False

        # MCH fru_start command expects fru number, instead of AMC
        # number, where:
        #     fru_num = amc_num + 4
        fru_num = amc_num + 4

        # Check current state of the AMC, and leave Telnet session
        # open
        res, state = self._get_slot_state(amc_num, False)

        # If unable to get state of slot, return
        if res is False:
            self.close()
            return False

        if state == "M4":
            self.logger.logger.info("AMC{} is already powered up.".format(amc_num))
            success = True
        else:
            # Sleep to allow previous telnet connection to properly
            # close
            time.sleep(1)

            if self._session_open is False:
                self.open()

            self.logger.logger.info("Powering up AMC {}.".format(amc_num))
            # Send fru_start command
            self._send_command("fru_start {}".format(fru_num))
            fru_start = self._read_command()

            # Check if the slot has a start up delay
            if upstream or ("Upstream Power Up Delay active!" in fru_start):
                self.logger.logger.info(
                    "Slot has upstream delay set. Waiting 60s for AMC to power on..."
                )
                sleep = 65

            # Sleep to allow device in AMC slot to power on
            time.sleep(sleep)

            # Get updated state
            _, state = self._get_slot_state(amc_num)

            if state == "M4":
                success = True
                self.logger.logger.info(
                    "AMC {} powered up successfully.".format(amc_num)
                )
            else:
                self.logger.logger.warning(
                    "Failed to power on slot. Current state: {}.".format(state)
                )

        self.close()

        return success

    def _stop_slot(self, amc_num, sleep: int = 2):
        """Internal method to stop the card in an AMC slot

        Args:
            amc_num(int): AMC slot number to power off.
            sleep(int): Time to allow for the AMC to power down.
                        Default: 2 seconds

        Returns:
            True if successful
            False if unsuccessful
        """
        success = False

        # MCH shutdown command expects fru number, instead of AMC
        # number, where:
        #     fru_num = amc_num + 4
        fru_num = amc_num + 4

        # Check current state of the AMC, and leave Telent
        # session open
        res, state = self._get_slot_state(amc_num, False)
        # If unable to get state of slot, return
        if res is False:
            self.logger.logger.warning(
                "Unable to determine state of slot {}:\n  {}".format(amc_num, state)
            )
            self.close()
            return False

        if state == "M1":
            self.logger.logger.info("AMC{} is already powered down.".format(amc_num))
            success = True
        else:

            # Sleep to allow previous telnet connection to properly
            # close
            time.sleep(1)

            if self._session_open is False:
                self.open()

            self.logger.logger.info("Powering down AMC {}.".format(amc_num))
            # Send shutdown command
            self._send_command("shutdown {}".format(fru_num))

            # Sleep to allow device in AMC slot to power down
            self.logger.logger.info("Sleeping for {} seconds.".format(sleep))
            time.sleep(sleep)

            # Get updated state
            _, state = self._get_slot_state(amc_num)

            if state == "M1":
                success = True
                self.logger.logger.info(
                    "AMC {} powered down successfully.".format(amc_num)
                )
            else:
                self.logger.logger.warning(
                    "Failed to power down slot. Current state: {}.".format(state)
                )

        self.close()

        return success

    def _slot_status_all(self):
        """Internal method to get the current state of all
        AMC slots.

        Args:
            close_conn(bool): Whether to close the telnet
               session when returning, or not.
               Default: True

        Returns:
            A list of tuples containing the AMC number and state code {M1, M2, ... M7} for all slots.
            If a slot is empty, the entry in the list will contain None.
        """
        self.logger.logger.debug("Getting state of AMCs in chassis.")

        if self._session_open is False:
            self.open()

        # Get fru_info for all AMCs
        self._send_command("show_fru")
        fru_info = self._read_command()

        self.close()

        # Search for populated AMC slots
        match = re.findall(r"AMC(\d+)(\s+)(M\d+)", fru_info)

        amcs = []

        # Create list for all AMCs
        for x in range(1, int(match[-1][0]) + 1):
            slot_state = [item for item in match if str(x) in item]
            if slot_state:
                self.logger.logger.debug(
                    "AMC{} slot is populated, and in the {} state.".format(
                        x, slot_state[0][2]
                    )
                )
                amcs.append((x, slot_state[0][2]))
            else:
                self.logger.logger.debug("AMC{} slot is unpopulated.".format(x))
                amcs.append((x, None))

        return amcs

    def _stop_all_slots(self):
        """Internal method to power down all AMC slots"""

        if self._session_open is False:
            self.open()

        self.logger.logger.info("Powering down all AMCs.")

        # Shutdown all slots
        self._send_command("shutdown all")

        self.logger.logger.info("All AMCs are now powered down.")

        self.close()

    def _determine_upstream(self):
        """Internal method to determine which AMCs are configured
        as 'Upstream', i.e the PCIe root-complex
        """
        if self._session_open is False:
            self.open()

        self.logger.logger.info("Determining which AMC(s) are configured as Upstream")

        # 'mch' command takes a while to return all info
        # so add an extended sleep period
        self._send_command("mch", sleep=2)
        mch_info = self._read_command()

        self.close()

        # Find the relevant lines in the output
        first = mch_info.find("VS #")
        last = mch_info.find("\r\nUpstream")
        pcie_info = mch_info[first:last].splitlines()

        """ Example info:
            ['  VS #  | Host    | NT-Host | Members',
             '  0       AMC01_4   none      AMC01_4 AMC02_4 AMC03_4 AMC04_4 AMC05_4 AMC06_4 AMC07_4 AMC08_4 AMC09_4 AMC10_4 AMC11_4 AMC12_4 ',
             '  1     ',
             '  2     ',
             '  3     ',
             '  4     ',
             '  5     ']
        """
        upstreams = []
        # There are upto 6  virtual switches
        for i in range(1, len(pcie_info)):
            # First line is the header
            match = re.findall(r"{}(\s+)AMC(\d+)_(\d)".format(i - 1), pcie_info[i])
            if len(match) > 0:
                self.logger.logger.info(
                    "AMC{} is the Upstream for Virtual Switch {}".format(
                        int(match[0][1]), i - 1
                    )
                )
                upstreams.append(int(match[0][1]))

        return upstreams

    def reboot_slot(self, amc_num):
        """ Method to power cycle an individual AMC slot.

        Args:
            amc_num: AMC slot number to power cycle.

        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, an empty string.
            - On failure, a string containing the failure message.
        """
        self.logger.logger.info("Rebooting AMC in slot {}".format(amc_num))

        response = ""

        # Sleep time
        shutdown_sleep = 5
        start_sleep = 5

        # Shutdown slot
        success = self._stop_slot(amc_num, shutdown_sleep)
        if success is False:
            response = "Failed to power down card in AMC slot {}.".format(amc_num)
            self.logger.logger.warning(response)
            return (success, response)

        # Allow time for session to update
        time.sleep(1)

        # Start slot
        success = self._start_slot(amc_num, start_sleep)
        if success is False:
            response = "Failed to power up card in AMC slot {}.".format(amc_num)
            self.logger.logger.warning(response)

        return (success, response)

    def reboot_all_slots(self):
        """ Method to power cycle all AMC slots.

        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, an empty string.
            - On failure, a string containing the failure message.
        """

        if self._session_open is False:
            self.open()

        # Find which slots are populated
        slot_status = self._slot_status_all()

        # Shutdown all slots
        self._stop_all_slots()

        # Determine what AMC(s) are the Upstream (root-complex)
        upstreams = self._determine_upstream()
        upstream_state = []

        overall_success = True
        overall_response = ""
        # Power on relevant slots
        for slot in slot_status:
            if (slot[1] is not None) and (slot[1] != "M1"):
                if slot[0] not in upstreams:
                    success = self._start_slot(slot[0])
                    if success is False:
                        response = "Failed to power up card in AMC slot {}.".format(
                            slot[0]
                        )
                        self.logger.logger.warning(response)
                        overall_success = False
                        overall_response += "{}\n".format(response)
                else:
                    upstream_state.append(slot)
                    self.logger.logger.info(
                        "AMC{} is configured as an upstream. Will reboot last.".format(
                            slot[0]
                        )
                    )

        # Power on Upstream slots
        for slot in upstream_state:
            if (slot[1] is not None) and (slot[1] != "M1"):
                success = self._start_slot(slot[0], upstream=True)
                if success is False:
                    response = "Failed to power up card in AMC slot {}.".format(slot[0])
                    self.logger.logger.warning(response)
                    overall_success = False
                    overall_response += "{}\n".format(response)

        self.close()

        return (overall_success, overall_response)



# -*- coding: utf-8 -*-

"""
gendev_interface
~~~~~~~~~~~~~~~~

Public API offering some useful features for configuring/maintaining devices
supported by the ICSHWI WP4.

This is just an interface, which means it has to be implemented for a
particular device. Don't use this directly, look inside the project for a
module that implements this interface.

Main features supported by the API:
- FW update
- Accessing/Setting configuration parameters
- Running dry checks against a golden configuration
"""

from __future__ import annotations
import abc
import enum
from logging import Logger
from collections import OrderedDict

__author__ = "Felipe Torres González"
__copyright__ = "Copyright 2021, ESS MCH Tools"
__credits__ = ["Felipe Torres González", "Ross Elliot", "Jeong Han Lee"]
__license__ = "GPL-3.0"
__version__ = "1.1"
__maintainer__ = "Ross Elliot"
__email__ = "ross.elliot@ess.eu"
__status__ = "Development"


class ConnType(enum.Enum):
    """This enumeration specifies the allowed connection types.

    Any device implementing the GenDevInterface should allow using one or more
    of the following connection types:

    - ETHER: Ethernet connection using TCP/IP.
    - SERIAL: Serial connection.
    - MOXA: Serial connection through a MOXA Hub (accessible over Ethernet).
    - TELNET: Serial connection through Ethernet using Telnet.
    - SSH: Secure Shell connection through Ethernet.
    """

    ETHER = 0
    SERIAL = 1
    MOXA = 2
    TELNET = 3
    SSH = 4


class GenDevInterface(metaclass=abc.ABCMeta):
    """Interface for a Generic Device.

    This interface can be implemented to control any device that includes some
    sort of management interface: command line trough a serial port, a web
    interface, ...

    Details for particular devices should be handled on their implementation
    for this interface.
    """

    def __init__(
        self,
        device_model: str,
        manufacturer: str,
        serial_num: str,
        allowed_conn: list,
        hostname: str = None,
        vlan: str = None,
        mac_address: str = None,
        logger: Logger = None,
        log_en: bool = False,
    ):
        """Class constructor.

        Args:
            device_model: The device identifier
            manufacturer: The device manufacturer
            serial_num: Serial number of the device
            allowed_conn: A list containing the supported connection types for
                         the device. See the ConnType enum for details.
            hostname: given hostname by CSEntry.
            vlan: vlan in which the device is registered.
            mac_address: MAC address of the network interface connected to the
                         TN network.
            logger: Reference to a logger instance. When this is not None, the
                    code will add info to the main's program log.
        """
        pass

    @abc.abstractmethod
    def device_info(self) -> tuple[bool, dict]:
        """Retrieve the main information about the device.

        The information is returned in a dictionary with 2 categories:
        *board* and *network*.

        Returns:
            If success, a dictionary with the device information.
            If failure, an empty dictionary on failure.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_dhcp_mode(self) -> tuple[bool, str]:
        """Enables DHCP mode in the network configuration of the device.

        Raises:
            ConnectionError: If the device is not accessible.
            NoValidConn: If no valid connection types supporting this feature
                         are used by the device.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update_fw(self, fw_version: str, part: str) -> tuple[bool, str]:
        """Update the firmware of the device.

        Args:
            fw_version: version release number for the new fw.
            part: modifier allowing the update of different parts within
                  the same device.

        Returns:
            If failure, it returns a tuple containing False, and a message
            about the failure.
            If success, it returns (True,)

        Raises:
            ConnectionError: If the device is not accessible.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_configuration(self, category: str, data, verify=True) -> tuple[bool, str]:
        """Change the configuration of the device.

        This method focuses on the configuration parameters that are not
        defined within a configuration script. Specifying the entire set of
        parameters is not mandatory, and also, a particular category of
        settings can be modified without affecting the rest.

        Args:
            category(str): settings category to be affected.
            data(dict): dictionary containing the values to be modified
            verify(bool): when True, the method performs a checking after
                          setting the new parameters.

        Returns:
            - 0 when successful or verify=False
            - A dictionary containing the values that are not matching the
              expectation. For each key, the expected and the given values are
              provided.

        Raises:
            ConnectionError: If the device is not accessible.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_configuration(self, category: str) -> tuple[bool, dict]:
        """Get the configuration of the device.

        This method returns a dictionary containing the configuration
        parameters of the device implementing this interface. If the device
        has several configuration categories, keys from the dictionary might
        contain other dictionaries.
        This method can be used for checking the good configuration of a
        device.

        Args:
            category: points to a subset of the configuration parameters of
                      the device.

        Returns:
            A dictionary containing the configuration of the device.

        Raises:
            ConnectionError: If the device is not accessible.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def check_configuration(
        self, category: str, config: OrderedDict
    ) -> tuple[bool, str]:
        """Check the settings of the device.

        This method retrieves the configuration parameters of a device,
        and performs a comparison against the values given by `config`.
        When differences are detected, the values are reported to a log,
        the method only returns a general message indicating the result
        of the check.

        Args:
            category (str): a key indicating the target settings when
            a device has several subset of configuration parameters.
            config (OrderedDict): a dictionary containing the expected
            values for all the parameters that are aimed to check.

        Returns:
            tuple[bool, str]: a boolean value indiciating whether the
            checking was successful or not; and a text message when the
            checking was unsuccesful.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _reboot(self, sleep: int = 50):
        """Internal method to reboot the MCH after a timeout.

        Args:
            timeout: Number of seconds to wait before rebooting the device.

        Raises:
            ConnectionError: If the device is not accessible.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _parse_config(self):
        """Internal method to parse the configuration of the MCH.

        When pulling content from the MCH, it might be needed to clean it and
        structure it properly so the rest of the logic can use the data
        efficiently or it can be presented to an user in a meaningful way.

        Raises:
            ConnectionError: If the device is not accessible.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def reboot_slot(self, amc_num) -> tuple[bool, str]:
        """ Method to power cycle an individual AMC slot.

        Args:
            amc_num: AMC slot number to power cycle.

        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, an empty string.
            - On failure, a string containing the failure message.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""
gendev_logger.py
~~~~~~~~~~

Contains helper methods for logging.
"""

from __future__ import annotations
import logging
from os import path, mkdir
from datetime import datetime

__author__ = "Ross Elliot"
__copyright__ = "Copyright 2021, ESS MTCA Tools"
__credits__ = ["Ross Elliot", "Felipe Torres González", "Jeong Han Lee"]
__license__ = "GPL-3.0"
__version__ = "1.1"
__maintainer__ = "Ross Elliot"
__email__ = "ross.elliot@ess.eu"
__status__ = "Development"


class GenDevLogger:
    def __init__(
        self,
        logger: logging.Logger = None,
        logname: str = None,
        logdir: str = "./logs",
        format: str = "%(asctime)s - %(levelname)s - %(message)s",
        log_level: int = logging.INFO,
        log_to_file: bool = False,
    ):
        """Class constructor.

        Args:
            logger(Logger): reference to an external Logger
            object to be used.
            logname(str): custom string for the log name.
            logdir(str): path to log directory.
            format(str): custom string defining the Logger
            format.
            log_level(int): log level to use.
        """

        self.format = format
        self.log_level = log_level
        self.logdir = logdir
        self.log_to_file = log_to_file

        # If no logname is provided, generate
        # the default one
        if logname is None:
            self.logname = self.genLogName(self.logdir)
        else:
            self.logname = self.logdir + "/" + logname

        # Specified logdir must exist
        if self.log_to_file:
            if not path.isdir(logdir):
                mkdir(logdir)

        # If a Logger object is provided, use it
        if logger is not None:
            self.logger = logger
            self._formatter = None
        else:
            # Create Logger object, and configure
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(logging.DEBUG)
            # Set default format
            self.setDefaultFormat()

            # Add default stream handlers
            if (not self.logger.hasHandlers()) or (len(self.logger.handlers) == 0):
                self.addDefaultHandlers(filename=self.logname)

    def setDefaultFormat(self):
        """Create a logging.Formatter instance, and provide
        the formatter string.
        """

        self._formatter = logging.Formatter(self.format)

    def addDefaultHandlers(self, filename):
        """Method to add default logging stream
           handlers to the Logger object.

        The default handlers are:
         - logging.StreamHandler(), for logging to
           the console.
         - logging.FileHandler, for logging to a file.

        Args:
            filename(str): name for the generated
            log file.
        """

        # Create handlers (console and file)
        h_console = logging.StreamHandler()

        if self.log_to_file:
            h_file = logging.FileHandler(filename)
            h_file.setLevel(logging.DEBUG)

        # If there is a defined format string, use it
        if self._formatter is not None:
            h_console.setFormatter(self._formatter)
            if self.log_to_file:
                h_file.setFormatter(self._formatter)

        h_console.setLevel(logging.INFO)

        # Add handlers to Logger object
        self.logger.addHandler(h_console)
        if self.log_to_file:
            self.logger.addHandler(h_file)

    def genLogName(self, dirname):
        """Generate a unique name for the log.'

        Will be of the ISO date format:

            ''YYYY_MM_DD_HHMMSS.txt''
        """

        # Get today's date/time in datetime format
        today = datetime.today()

        # Form log name
        logname = "{}/log_{}{:02}{:02}_{:02}{:02}{:02}.txt".format(
            dirname,
            today.year,
            today.month,
            today.day,
            today.hour,
            today.minute,
            today.second,
        )

        return logname



# -*- coding: utf-8 -*-

"""
nat_mch.py
~~~~~~~~~~

Implementation of the GenDev interface for NAT MCHs.

This module fully implements the GenDev interface. This is achieved by the use
of several sub-modules which implement different ways to access an NAT MCH.
This device doesn't allow a simple and straightforward implementation using an
only interface, so that, multiple communication ways are used. The golden rule
is achieving a reliable solution with the best performance.
"""

from __future__ import annotations
import logging
from pprint import pformat
from enum import IntEnum
from collections import OrderedDict
from difflib import Differ
from ..gendev_interface import GenDevInterface, ConnType
from ..gendev_err import ConnNotImplemented, FeatureNotSupported, DHCPEnableFailed
from ..gendev_logger import GenDevLogger
from .nat_mch_web import NATMCHWeb
from .nat_mch_telnet import NATMCHTelnet
from .nat_mch_moxa import NATMCHMoxa

__author__ = "Felipe Torres González"
__copyright__ = "Copyright 2021, ESS MCH Tools"
__credits__ = ["Felipe Torres González", "Ross Elliot", "Jeong Han Lee"]
__license__ = "GPL-3.0"
__version__ = "1.1"
__maintainer__ = "Ross Elliot"
__email__ = "ross.elliot@ess.eu"
__status__ = "Development"


class BackplaneType(IntEnum):
    """Enum defining the types of backplane for mTCA crates."""

    B1U = 1
    B3U = 3
    B5U = 5
    B9U = 9


class NATMCH(GenDevInterface):
    """NAT MCH device.

    This class is connection-agnostic, which means there are no internal
    details about the particular implementation attending to the chosen
    connection type. This class should call child subclasses depending on
    the chosen connection type and format the information properly.

    When two interfaces could be used for the same purpose, the most reliable
    should be used. That is auto managed by the class when all the supported
    communication interfaces are properly specified to the module.

    The order in which the communication interfaces are prioritized is:
    `Web >> SSH >> Telnet >> MOXA >> Serial`

    Within the interface module, an enumerated type is provided defining all
    the allowed communication methods. Specify what are supported by the
    MCH when instantiating this class. As a rule of thumb, if the MCH is able
    to get a valid IP address, all the interfaces relying on the network are
    available, the regular ETHER communication type is the best option, as it
    directly access the resources from the web server of the MCH (fastest and
    most reliable method), but ETHER doesn't support the execution of all the
    methods offered by the API, check the documentation of each method to
    check what type of connection is required.
    """

    def __init__(
        self,
        ip_address: str,
        allowed_conn: list[ConnType],
        backplane: BackplaneType,
        device_model: str = "MCH",
        manufacturer: str = "NAT",
        hostname: str = None,
        conn_ip_address: str = None,
        conn_port: int = None,
        logger: logging.Logger = None,
        log_to_file: bool = False,
    ):
        """Class constructor.

        Initialization of an MCH class.In order to enable logging, specify
        a valid reference to a Logger instance.

        Args:
            ip_addr(str): the given IP to the MCH in CSEntry.
            allowed_conn(list[ConnType]): list of connections supported by the MCH.
            backplane(BackplaneType): type of backplane of the crate in which the MCH is installed.
            device_model(str): string identifying the device.
            manufacturer(str): name of the manufacturer.
            hostname(str): the registered hostname in CSEntry for the MCH.
            conn_ip_address(str): ip address used for the communication with the MCH (MOXA backend)
            conn_port(str): port used for the communication with the MCH (MOXA backend).
            logger: reference to a Logger instance.

        Raises:
            ConnNotImplemented: if a communication interface that
            is not supported by the implementation was included in the
            *allowed_con* argument.
        """
        self.ip_address = ip_address
        self.device_model = device_model
        self.manufacturer = manufacturer
        self.serial_num = ""
        self.fw_ver = ""
        self.allowed_conn = allowed_conn
        self.hostname = ""
        self.mac_address = ""
        self.backplane = backplane
        self._conn_ip_address = conn_ip_address
        self._conn_port = conn_port
        self._log_to_file = log_to_file

        # Create a GenDevLogger instance from Logger object
        self.logger = GenDevLogger(logger=logger, log_to_file=self._log_to_file)
        self.logger.logger.info(
            "GenDev::Constructor - A new device has been registered"
            "\tDevice model: {}".format(self.device_model)
        )

        if log_to_file:
            self.logger.logger.info(
                "Log file located at: {}".format(self.logger.logname)
            )

        # Open the valid connections
        if ConnType.ETHER in self.allowed_conn:
            self._eth_conn = NATMCHWeb(
                self.ip_address,
                logger=self.logger,
            )
        if ConnType.TELNET in self.allowed_conn:
            self._tel_conn = NATMCHTelnet(
                ip_address=self.ip_address,
                hostname=self.hostname,
                backplane=self.backplane,
            )
        if ConnType.SERIAL in self.allowed_conn:
            raise ConnNotImplemented(
                "The serial interface is not implemented" " for NAT MCHs."
            )
        if ConnType.MOXA in self.allowed_conn:
            self._mox_conn = NATMCHMoxa(
                mch_ip_address=self.ip_address,
                moxa_ip_address=self._conn_ip_address,
                port=self._conn_port,
                backplane=self.backplane,
                logger=self.logger,
                hostname=self.hostname,
            )
        if ConnType.SSH in self.allowed_conn:
            raise ConnNotImplemented(
                "The SSH interface is not implemented" " for NAT MCHs."
            )

        # This dict holds a look-up table to translate argument keys, given to the
        # methods of this class, to keys from the configuration files.
        self._cfgkeys = {
            "basecfg": {
                "cfgkey": "Base MCH parameter",
            },
            "pcie": {
                "cfgkey": "PCIe parameter",
            },
            "backplane": {"cfgkey": "Backplane configuration"},
        }

    def device_info(self) -> tuple[bool, dict]:
        """Retrieve the main information about the device.

        The information is returned in a dictionary with 2 categories:
        *Board* and *Network*.
        This feature is supported by all the implemented communication
        interfaces, so the best one is chosen when multiple are allowed.

        An example of the returned dictionary::

            {
              'Board': {
                'fw_ver': 'V2.21.8',
                'fpga_ver': 'V1.14',
                'mcu_ver': '1.2',
                'serial_num': '113522-1426'},
              'Network': {'ip_address': '172.30.5.238',
                'mac_address': '00:40:42:22:05:92',
                'subnet_address': '255.255.252.0',
                'gateway_address': '172.30.7.254'}
            }


        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, a dictionary with the device information.
            - On failure, an empty dictionary.

        Raises:
            FeatureNotSupported: if the given allowed communication \
            interfaces don't allow running this method.
            oRouteToDevice: when the device is not reachable in the network, or the \
            device pointed by the given IP address is not an MCH.
        """
        if ConnType.ETHER in self.allowed_conn:
            valid, response = self._eth_conn.device_info()
        elif ConnType.TELNET in self.allowed_conn:
            valid, response = self._tel_conn.device_info()
        elif ConnType.MOXA in self.allowed_conn:
            valid, response = self._mox_conn.device_info()
        else:
            msg = (
                "Impossible to retrieve the device "
                "information with the given allowed "
                "communication interfaces to the MCH."
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        if valid:
            self.mac_address = response["Network"]["mac_address"]
            self.serial_num = response["Board"]["serial_num"]
            self.fw_ver = response["Board"]["fw_ver"]
            self.logger.logger.info(
                "\n[Board info]\n \
                   Serial number:    {}\n \
                   FPGA version:     {}\n \
                   Firmware version: {}\n \
                   MCU Version:      {}\n".format(
                    self.serial_num,
                    response["Board"]["fpga_ver"],
                    response["Board"]["fw_ver"],
                    response["Board"]["mcu_ver"],
                )
            )
            self.logger.logger.info(
                "\n[Network info]\n \
                   IP Address:  {}\n \
                   MAC Address: {}\n \
                   Subnet:      {}\n \
                   Gateway:     {}\n".format(
                    self.ip_address,
                    self.mac_address,
                    response["Network"]["subnet_address"],
                    response["Network"]["gateway_address"],
                )
            )

        return valid, response

    def set_dhcp_mode(self) -> tuple[bool, str]:
        """Enables DHCP mode in the network configuration of the device.

        Performs the following steps:
            - Enables DHCP mode on the MCH.
            - Sets the internal IP address value to match the address
              provided by the DHCP server(required to prevent DHCP lease
              issues).
            - Sets the hostname value.

        Returns
            If failure, a tuple containing False, and a message about the
            failure.
            If success, a tuple containing True, and an empty string.

        Raises:
            DHCPEnableFailed: If the setting fails.
            NoValidConn: If no valid connection types supporting this feature
                         are used by the device.
        """
        if ConnType.TELNET in self.allowed_conn:
            success, response = self._tel_conn.set_dhcp_mode()
        elif ConnType.MOXA in self.allowed_conn:
            success, response = self._mox_conn.set_dhcp_mode()
        else:
            msg = (
                "Setting the DHCP mode is not supported for the given "
                "connection type"
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        if not success:
            self.logger.logger.error(response)
            raise DHCPEnableFailed(response)

        return success, response

    def update_fw(self, fw_version: str, part: str = None) -> tuple[bool, str]:
        """Update the firmware of the device.

        This feature is only supported by the Telnet communication interface.

        This method expects the firmware binary pointed by the value of the
        argument *fw_version* to be available in the TFTP server.
        Mainly, this method injects the command *update_firmware* to an NAT
        MCH.

        Args:
            fw_version: version release number for the new fw.
            part: modifier allowing the update of different parts within
                  the same device.

        Returns:
            If failure, it returns a tuple containing False, and a message
            about the failure.
            If success, it returns (True,)

        Raises:
            ConnectionError: If the device is not accessible.
        """
        if ConnType.TELNET in self.allowed_conn:
            valid, response = self._tel_conn.update_fw(fw_version)
        elif ConnType.MOXA in self.allowed_conn:
            valid, response = self._mox_conn.update_fw(fw_version)
        else:
            msg = (
                "Impossible to update the fw of the "
                "device with the given allowed "
                "communication interfaces to the MCH."
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        if valid:
            self.logger.logger.info(response)
        else:
            self.logger.logger.error(response)

        return valid, response

    def set_configuration(
        self, category: str, data: OrderedDict, apply: bool = True
    ) -> tuple[bool, str]:
        """Change the configuration of the device.

        This method focuses on the configuration parameters that are not
        defined within a configuration script. Specifying the entire set of
        parameters is not mandatory, and also, a particular category of
        settings can be modified without affecting the rest.

        This method supports the following configuration categories (taken
        from the webpage names):

        - Base Configuration [basecfg]
        - PCIe Virtual Switches [pcie]
        - Backplane configuration [backplane]

        Args:
            category(str): the target set of parameters to be affected by the change. \
            The accepted values are previously listed.
            data(dic): dictionary containing the values to be modified.
            apply(bool): whether to reboot the MCH to make active the changes or not. \
            Disable rebooting when multiple changes are going to be performed in the device. \
            Reboot  the device after the last change.

        Returns:
            A tuple containing a `bool` indicating if the operation was successful,
            and an error message.

        Raises:
            ConnectionError: If the device is not accessible.
            NoRouteToDevice: when the device is not reachable after applying the \
            new configuration.
        """
        if ConnType.ETHER in self.allowed_conn and category != "backplane":
            valid, response = self._eth_conn.set_configuration(category, data, apply)
        elif ConnType.TELNET in self.allowed_conn and category == "backplane":
            valid, response = self._tel_conn.set_configuration(category, data, apply)
        elif ConnType.MOXA in self.allowed_conn and category == "backplane":
            valid, response = self._mox_conn.set_configuration(category, data, apply)
        else:
            msg = (
                "Setting the MCH configuration using the "
                "{} backend is not implemented.".format(str(self.allowed_conn))
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        if not valid:
            self.logger.logger.error(response)

        return valid, response

    def get_configuration(self, category: str = None) -> tuple[bool, OrderedDict]:
        """Get the configuration of the device.

        This method returns a dictionary containing the configuration
        parameters of the device implementing this interface. If the device
        has several configuration categories, keys from the dictionary might
        contain other dictionaries.

        This method supports the following configuration categories (taken
        from the webpage names):

        - Base Configuration [basecfg]
        - PCIe Virtual Switch Configuration [pcie]
        - Backplane Configuration [backplane]

        Args:
            category(str): points to a subset of the configuration parameters of \
            the device. Use the values given between brackets from the previous \
            item list.

        Returns:
            tuple[bool, OrderedDict]: The first element indicates the success of \
            the operation, the second is an ordered dict containing the settings in \
            the same order as they are found in the web page. When *category* = \
            **backplane**, the dictionary contains only one key (*Backplane \
            config file*) and the whole configuration file as value for that \
            key.
        """
        if ConnType.ETHER in self.allowed_conn:
            valid, response = self._eth_conn.get_configuration(category)
        else:
            msg = (
                "Retrieving the MCH configuration using the "
                "{} backend is not implemented.".format(str(self.allowed_conn))
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        if valid:
            self.logger.logger.debug(pformat(response))
        else:
            self.logger.logger.error(response)

        return valid, response

    def check_configuration(
        self, category: str, config: OrderedDict
    ) -> tuple[bool, str]:
        """Check the settings of the device.

        This method retrieves the configuration parameters of a device,
        and performs a comparison against the values given by `config`.
        When differences are detected, the values are reported to a log,
        the method only returns a general message indicating the result
        of the check.

        Args:
            category (str): a key indicating the target settings when
            a device has several subset of configuration parameters.
            config (OrderedDict): a dictionary containing the expected
            values for all the parameters that are aimed to check.

        Returns:
            tuple[bool, str]: a boolean value indiciating whether the
            checking was successful or not; and a text message when the
            checking was unsuccesful.
        """
        if ConnType.ETHER not in self.allowed_conn:
            msg = (
                "Accessing the MCH configuration is only possible "
                "when using the Web backend."
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        # Is `config` good?
        try:
            valid_args = self._eth_conn._check_arguments(config, category)
            if not valid_args[0]:
                self.logger.logger.info(
                    "Configuration check for {} passed successfully.".format(category)
                )
                return valid_args
        except KeyError:
            msg = "The given configuration set contains unknown parameters."
            self.logger.logger.error(msg)
            return (False, msg)

        valid, mch_config = self._eth_conn.get_configuration(category)

        if not valid:
            self.logger.logger.info(
                "Configuration check for {} failed.".format(category)
            )
            return valid, mch_config

        key = self._cfgkeys[category]["cfgkey"]

        # The backplane configuration is a special thing
        if category == "backplane":
            # The initial comment lines are stripped from the golden config
            # when it is applied to the MCH.
            #
            # Strip the first 11 lines before comparing

            # Split string by newline character
            mch_golden_conf_list = config[key]["file"].splitlines(keepends=True)
            mch_conf_list = mch_config[key]["file"].splitlines(keepends=True)

            # Create Differ() object
            differ = Differ()
            diff = list(differ.compare(mch_conf_list, mch_golden_conf_list))

            if mch_config[key]["file"] == config[key]["file"]:
                ret = (True, "")
            else:
                ret = (
                    False,
                    "Backplane configuration setting failed. See log for details.",
                )
                self.logger.logger.debug(
                    "Backplane configuration does not match golden config. Diff provided:"
                )
                self.logger.logger.debug(pformat(diff))

            return ret

        # Fast comparison: keys
        if mch_config[key].keys() != config[key].keys():
            msg = "The given set of parameters doesn't match the values from the MCH."
            self.logger.logger.error(msg)
            return (False, msg)

        # Second, compare the content
        result = (True, "")
        skip_keys = []
        if category == "basecfg":
            skip_keys = ["DHCP parameter"]
        try:
            result = self._parse_config(mch_config, config, key, skip_keys)
        except KeyError as e:
            # Add something to the log
            msg = "The configuration is missing a key: {}".format(e)
            self.logger.logger.error(msg)
            result = (False, msg)

        if not result[0]:
            self.logger.logger.error(
                "Configuration check for {} failed.".format(category)
            )
        else:
            self.logger.logger.info(
                "Configuration check for {} succeeded.".format(category)
            )

        return result

    def _parse_config(
        self, dict1: OrderedDict, dict2: OrderedDict, key, skip_keys: list = []
    ) -> tuple[bool, str]:
        """Compare two dictionaries, item by item.

        This method iterates over the dictionaries by levels. It goes to the bottom
        level of the hierarchy for the given dictionaries. Then it compares all the
        values for all the keys between both dictionaries. If no differences are
        found, it goes a level up, and repeats the operation until reaching the root
        key of the dictionaries.

        Pre:
            Both dictionaries should contain the same keys following the same
            hierarchy structure.

        Args:
            dict1 (OrderedDict): Main dictionary. This one is used to extract the keys. \
            Only use trusted dictionaries for this argument.
            dict2 (OrderedDict): Slave dictionary. This one is accessed using the keys from \
            `dict1`. If using untrusted content, like coming from an user, introduce it using \
            this argument.
            skip_keys (list): List of keys to skip during checks.
        Returns:
            tuple[bool, str]: Success status and error message when a difference was found.

        Raises:
            KeyError: When `dict2` doesn't contain the same keys in the same levels than `dict1`.

        """
        for k in dict1[key]:
            if isinstance(dict1[key][k], OrderedDict):
                if k in skip_keys:
                    self.logger.logger.debug('Skipping check for "{}"'.format(k))
                    continue
                else:
                    self.logger.logger.debug("{}:".format(k))
                    check, msg = self._parse_config(dict1[key], dict2[key], k)
                    if not check:
                        return (check, msg)
            elif dict1[key][k] != dict2[key][k]:
                # DHCP hostname cannot be checked, as it differs for each
                # MCH
                if k not in skip_keys:
                    self.logger.logger.error(
                        "{} != {}".format(dict1[key][k], dict2[key][k])
                    )
                    msg = "The configuration check for {} failed.".format(k)
                    self.logger.logger.error(msg)
                    return (False, msg)
            else:
                self.logger.logger.debug(
                    "-- {}, {} == {}".format(k, dict1[key][k], dict2[key][k])
                )

        return (True, "")

    def _reboot(self, sleep: int = 60):
        """Internal method to reboot the MCH after a timeout.

        Args:
            sleep: Number of seconds to wait after rebooting the device.
        """
        if ConnType.ETHER in self.allowed_conn:
            self._eth_conn._reboot(sleep)
        elif ConnType.TELNET in self.allowed_conn:
            self._tel_conn._reboot(sleep)
        elif ConnType.MOXA in self.allowed_conn:
            self._mox_conn._reboot(sleep)
        else:
            msg = (
                "Impossible to reboot the device "
                "with the given allowed "
                "communication interfaces to the MCH."
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

    def reboot_slot(self, amc_num) -> tuple[bool, str]:
        """ Method to power cycle an individual AMC slot.

        Args:
            amc_num: AMC slot number to power cycle.

        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, an empty string.
            - On failure, a string containing the failure message.
        """
        if ConnType.TELNET in self.allowed_conn:
            success, response = self._tel_conn.reboot_slot(amc_num)
        elif ConnType.MOXA in self.allowed_conn:
            success, response = self._mox_conn.reboot_slot(amc_num)
        else:
            msg = (
                "The current connection to the MCH "
                "does not allow for power cycling "
                "of individual slots."
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        if not success:
            self.logger.logger.error(response)
        else:
            self.logger.logger.info(
                "Successfully rebooted AMC in slot {}.".format(amc_num)
            )

        return (success, response)

    def reboot_all_slots(self) -> tuple[bool, str]:
        """ Method to power cycle all AMC slots.

        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, an empty string.
            - On failure, a string containing the failure message.
        """
        if ConnType.TELNET in self.allowed_conn:
            success, response = self._tel_conn.reboot_all_slots()
        elif ConnType.MOXA in self.allowed_conn:
            success, response = self._mox_conn.reboot_all_slots()
        else:
            msg = (
                "The current connection to the MCH "
                "does not allow for power cycling "
                "of slots."
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)

        if not success:
            self.logger.logger.error(response)
        else:
            self.logger.logger.info("Successfully rebooted all AMC slots.")

        return (success, response)



# -*- coding: utf-8 -*-

"""
nat_mch_moxa.py
~~~~~~~~~~~~~~~

Communication via Moxa is performed via a telnet connection. Therefore,
the Moxa class is implemented as a child class that inherits from NATMCHTelnet.
As such, all operations supported by the telnet class are supported by the MOXA
class.
"""
import time
import re
from gendev_tools.nat_mch.nat_mch_telnet import NATMCHTelnet
from ..gendev_logger import GenDevLogger

__author__ = "Ross Elliot"
__copyright__ = "Copyright 2021, ESS MTCA Tools"
__credits__ = ["Felipe Torres González", "Jeong Han Lee"]
__license__ = "GPL-3.0"
__version__ = "1.1"
__maintainer__ = "Ross Elliot"
__email__ = "ross.elliot@ess.eu"
__status__ = "Development"


class NATMCHMoxa(NATMCHTelnet):
    """NATMCHMoxa access an NAT MCH via Moxa.

    This module implements some operations using the command line interface
    via Moxa. Thus, the MCH has to be connected to a Moxa serial server.
    The firmware update relies on **the mchconfig-server** to serve the
    firmware image using the FTP protocol.

    Supported operations:
    - Retrieve the general information of the MCH.
    - Firmware update of the MCH.
    - Enabling of DHCP on the MCH.
    """

    def __init__(
        self,
        mch_ip_address: str,
        moxa_ip_address: str,
        port: int,
        backplane,
        logger: GenDevLogger = None,
        hostname: str = None,
        **kwargs,
    ):
        """Class constructor.

        Args:
            ip_address: the IP address of the Moxa serial server.
            port: port of the Moxa server that the MCH is connected to, in the
                  range [0, 1, ... 16]
            logger: reference to a logger that is being used

        Raises:
            gendev_err.ConnTimeout if the device is not reachable.
        """

        # Validate Moxa port is within range
        if 0 <= port >= 16:
            raise ValueError("Moxa port must be in range of [0 to 16]")

        # Real Telnet port used to connect to Moxa is in range of [4000 t0 4016]
        self.port = 4000 + port
        self.moxa_ip_address = moxa_ip_address

        if "log_to_file" in kwargs:
            log_to_file = kwargs["log_to_file"]
        else:
            log_to_file = False

        # Initialise parent class and inherit methods and properties
        super().__init__(
            ip_address=mch_ip_address,
            backplane=backplane,
            port=self.port,
            logger=logger,
            hostname=hostname,
            telnet_ip_address=moxa_ip_address,
            log_to_file=log_to_file,
        )

    def _send_command(
        self,
        command: str,
        sleep: int = 1,
        clear_buffer: bool = True,
        chunk_size: int = 3,
    ):
        """Internal method for sending a low level command to the MCH via a Moxa
        server.

        This command allows forgetting about the particular details of using
        a Telnet session behind the scenes. A regular command from the MCH
        command line interface can be sent through this interface without
        worrying about the underlying communication.
        This implementation for the MOXA required splitting any command into
        chunks of 3-characters or less, to remedy an issue with the MOXA where
        some of the characters are lost during transmission.

        Args:
            command: command to be sent to the MCH.
            sleep: amount of seconds to wait after sending a command.
            clear_buffer: send a carriage return before the command. This
            helps clearing previous garbage from the buffer, but
            it should be used with caution because there are
            commands that doesn't expect a carriage return after.
            chunk_size: number of characters to use for each chunk when
            splitting the command string.
        """
        # clean up
        if clear_buffer:
            self._session.write(b"\r")
            self._session.read_until(b"nat> ")
            time.sleep(sleep)

        # Split command into 3-char chunks
        pattern = ".{{1,{chunk_size}}}".format(chunk_size=chunk_size)
        chunks = re.findall(pattern, command)
        # Write each chunk serially
        for chunk in chunks:
            self._session.write(chunk.encode("ascii"))
            time.sleep(sleep)
        # Send carriage return to issue command
        self._session.write(b"\r")
        time.sleep(sleep)



# -*- coding: utf-8 -*-
"""
nat_mch_web.py
~~~~~~~~~~~~~~

Incomplete implementation of the GenDev interface. The ultimate purpose
of this module is to be encapsulated within another module that really
implements that interface.

Unfortunately, NAT MCHs don't allow to go for a full implementation of the
GenDev interface based on an only communication interface. This module makes
use of the Web interface of the MCH for extracting data and change data when
possible.

This module aims to provide a fast but reliable mechanism to change the
configuration parameters of the MCH. Rely on this implementation for
interfacing an NAT MCH whenever possible.
"""

from __future__ import annotations
import re
import time
import requests as rq
from collections import OrderedDict
from bs4 import BeautifulSoup
from ..gendev_err import FeatureNotSupported, NoRouteToDevice, WebChanged
from ..gendev_logger import GenDevLogger

__author__ = ["Felipe Torres González", "Ross Elliot"]
__copyright__ = "Copyright 2021, ESS MCH Tools"
__license__ = "GPL-3.0"
__version__ = "1.1"
__maintainer__ = "Ross Elliot"
__email__ = "ross.elliot@ess.eu"
__status__ = "Release Candidate"


class NATMCHWeb:
    """NATMCHWeb access an NAT MCH via the web interface.

    This module makes use of the Web interface provided by NAT MCHs to access
    and change the values of the configuration. This interface is the most
    reliable among all the offered communication interfaces to an MCH. The
    only caveat is that the interface is not designed to be used by *sw* but an
    human. That means, some particular features are not easy to support or they
    seem impossible to do.
    The main requirement for using this module is having the Web interface
    enabled in the MCH configuration.

    Supported operations:
    - Retrieve the general information of the MCH.
    - Change/Access the general configuration of the MCH.
    - Change/Access the PCIe configuration of the MCH.
    - Access the backplane configuration of the MCH.
    - Check the backplane configuration of the mch against the content of a file.
    """

    def __init__(
        self,
        ip_address: str,
        logger: GenDevLogger = None,
        open: bool = True,
        **kwargs,
    ):
        """Class constructor.

        Args:
            ip_address(str): the IP address of the MCH.
            logger(Logger): instance of a system Logger. When this argument is set to \
            a valid object, this module will produce info/debug output to the logger.
            open(bool): whether to establish the connection at instantiation time or later. \
            Set this argument to False when a deferred link establishment is aimed.

        Raises:
            NoRouteToDevice: when the device is not reachable in the network or the IP \
            is used by a device that is not an NAT MCH.
        """
        self.ip_address = ip_address

        if "log_to_file" in kwargs:
            self._log_to_file = kwargs["log_to_file"]
        else:
            self._log_to_file = False

        # Header for the HTML methods, the most important variable is the
        # Authorization because NAT MCHs need to login using Root:NAT.
        self._http_headers = {
            "Accept-Language": "en-GB,en;q=0.5",
            "Authorization": "Basic cm9vdDpuYXQ=",
            "Connection": "keep-alive",
        }

        # Initialise Logger instance
        if isinstance(logger, (GenDevLogger)):
            self.logger = logger
        else:
            self.logger = GenDevLogger(logger=logger, log_to_file=self._log_to_file)

        self.logger.logger.info(
            "NATMCHWeb - NAT MCH Web instance created."
            "\tMCH IP Address: {}".format(self.ip_address)
        )

        # Flag indicating when the link is actually established
        #
        self._link_established = open
        if self._link_established:
            self.open()

        # Regular expressions for parsing data from the Board Information page
        self._match_fw_ver = re.compile(
            r"Firmware Version\n(V\d{1,2}\.\d{1,2}\.\d{1,2})"
        )
        self._match_fpga_ver = re.compile(r"FPGA Version\n(V\d{1,2}\.\d{1,2})")
        self._match_mcu_ver = re.compile(
            r"Microcontroller Version\n(V\d{1,2}\.\d{1,2})"
        )
        self._match_board_sn = re.compile(r"Board Serial Number\n(\d{6}-\d{4})")
        self._match_ip_addr = re.compile(r"IP Address\n((\d{1,3}\.?){4})")
        self._match_mac_addr = re.compile(r"IEEE Address\n(([\d\D]{2}:?){6})")
        self._match_subnet_mask = re.compile(r"Subnet Mask\n((\d{1,3}\.?){4})")
        self._match_gateway_addr = re.compile(r"Gateway Address\n((\d{1,3}\.?){4})")

        # This dict holds a look-up table to translate argument keys given to the
        # methods of this class, to keys from the configuration files and keys
        # for the HTML GET/POST requests.
        self._cfgkeys = {
            "basecfg": {
                "cfgkey": "Base MCH parameter",
                "webkey": ["get_mch_cfg"],
                "type": "dict",
            },
            "pcie": {
                "cfgkey": "PCIe parameter",
                "webkey": [
                    "pcie_width_link_ctrl",
                    "pcie_vs_cfg_refresh",
                    "pcie_vs_cfg_backup_flash_save",
                ],
                "type": "dict",
            },
            "backplane": {
                "cfgkey": "Backplane configuration",
                "webkey": ["web_cfg_backup_show_wait_verify"],
                "type": "dict",
            },
        }

    def open(self):
        """Establish the connection to the target MCH.

        Raises:
            NoRouteToDevice: when the device is not reachable in the network, or the \
            device pointed by the given IP address is not an MCH.
        """
        try:
            response = rq.get(
                "http://{}/index.asp".format(self.ip_address),
                headers=self._http_headers,
                timeout=2,
            )
        except Exception as e:
            if isinstance(e, rq.exceptions.ConnectionError):
                msg = "Error connecting to MCH web interface at {0}:".format(
                    self.ip_address
                )
                self.logger.logger.error(msg)
                raise NoRouteToDevice(msg)

        if not response.ok:
            msg = "Unable to reach device at {0}, status code {1}:".format(
                self.ip_address, response.status_code
            )
            self.logger.logger.error(msg)
            raise NoRouteToDevice(msg)
        else:
            html_content = BeautifulSoup(response.text, "lxml")
            title = html_content.head.title.text

            if not title == "MCH Configuration":
                msg = "Device at IP {0} is not an MCH. Retrieved description of device: ''{1}''".format(
                    self.ip_address, title
                )
                self.logger.logger.error(msg)
                raise NoRouteToDevice(msg)

        self._link_established = True

    def _parse_device_info(self, response) -> OrderedDict:
        """Internal method to parse the HTML content of the board information page.

        Args:
            response: The output from the requests.get call.

        Returns:
            An OrderedDict containing the settings for the baord information page \
            in the MCH web page.
        """
        html_content = BeautifulSoup(response.text, "html.parser")
        raw_info = html_content.get_text()
        resp_dict = OrderedDict()  # type: OrderedDict
        resp_dict["Board"] = OrderedDict()

        resp_dict["Board"]["fw_ver"] = self._match_fw_ver.search(raw_info).group(1)
        resp_dict["Board"]["fpga_ver"] = self._match_fpga_ver.search(raw_info).group(1)
        # The web interface returns the version number with the prefix 'V',
        # while the other interfaces have no prefix.
        # Remove the prefix for consistency
        resp_dict["Board"]["mcu_ver"] = (
            self._match_mcu_ver.search(raw_info).group(1).strip("V")
        )
        resp_dict["Board"]["serial_num"] = self._match_board_sn.search(raw_info).group(
            1
        )

        resp_dict["Network"] = OrderedDict()
        resp_dict["Network"]["ip_address"] = self._match_ip_addr.search(raw_info).group(
            1
        )
        resp_dict["Network"]["mac_address"] = self._match_mac_addr.search(
            raw_info
        ).group(1)
        resp_dict["Network"]["subnet_address"] = self._match_subnet_mask.search(
            raw_info
        ).group(1)
        resp_dict["Network"]["gateway_address"] = self._match_gateway_addr.search(
            raw_info
        ).group(1)

        return resp_dict

    def _parse_basecfg(self, response):
        """Internal method to parse the HTML content for the base configuration.

        Args:
            response: The output from the requests.get call.

        Returns:
            A OrderedDict containing the settings for the base configuration
            page in the MCH webpage.
        """
        mch_config = OrderedDict()
        # The Xilinx table is only parsed properly using the html5lib
        html_content = BeautifulSoup(response.text, "html5lib")
        tables = html_content.body.find_all("table")

        # Each subsection of the MCH Configuration is a different HTML
        # table. Each table will add a child dictionary to the main one,
        # indexed by the tr tag label from the HTML code.
        for table in tables:
            # Each section has a th tag. Then a set of parameters are included
            # within the section.
            table_title = table.th.text.strip()
            # Set the table heather as the dict key
            mch_config[table_title] = OrderedDict()

            for row in table.find_all("tr"):
                # Now, we're inside a section. Two kinds of parameters:
                # - parameters with a numeric value
                # - parameters with a select

                select_params = row.find("select")
                if select_params:
                    # Form with select values
                    # 1. First get the name attribute (should use find_all?)
                    name = select_params["name"].strip()
                    # 2. Look for the selected value for the given parameter
                    subrows = row.find_all("option")
                    value = [r["value"] for r in subrows if "selected" in str(r)][0]

                    # Rough check, but better than nothing
                    if name is None or value is None:
                        continue

                    mch_config[table_title][name] = value
                    # No more stuff to parse for this row
                    continue

                # Form with input values
                input_params = row.find_all("input")

                if not input_params:
                    continue

                # How many fields does this setting have?
                if len(input_params) > 1:
                    name = input_params[0]["name"]
                    value = [v["value"] for v in input_params]
                else:
                    name = input_params[0]["name"]
                    value = input_params[0]["value"]

                if name is None or value is None:
                    continue
                mch_config[table_title][name] = value

        # The previous set of parameters will be encapsulated so
        # a global dictionary can contain other set of parameters
        globalcfg = OrderedDict()
        globalcfg["Base MCH parameter"] = mch_config

        return globalcfg

    def _parse_pcie(self, response):
        """Internal method to parse the HTML content for the PCIe configuration.

        This method receives the content of the page /goform/pcie_vs_config* and
        extracts the information from the Link width and Virtual switch tables.

        Args:
            response: The output from the requests.get call.

        Returns:
            A OrderedDict containing the settings for the PCIe configuration
            page in the MCH webpage.
        """
        mch_config = OrderedDict()
        # Use lxml as parser, html.parser doesn't parse this web properly!
        html_content = BeautifulSoup(response.text, "lxml")
        # Divide the content by forms
        forms = html_content.body.find_all("form")

        # The first form contains the Link with configuration
        cfgtitle = "Link Width Configuration"
        if (
            forms[0].attrs["action"] != "/goform/pcie_width_link_ctrl"
            or forms[1].attrs["action"] != "/goform/pcie_vs_cfg_refresh"
        ):
            raise WebChanged("The PCIe configuration page has changed its format")

        # There're 3 tables, each one configures a pair of adjacent AMC slots.
        # From each table, get the value selecting a link with for each pair
        # of AMC slots.
        link_inputs = forms[0].find_all("input")
        mch_config[cfgtitle] = OrderedDict()
        curr_station = "Station_0"

        for input in link_inputs:
            if input["name"] == curr_station and "checked" in input.attrs:
                mch_config[cfgtitle][curr_station] = input["value"]
                next_station = int(curr_station[-1]) + 1
                curr_station = "Station_{}".format(next_station)

        # Now, extract the information from the table with the Virtual Switch
        # Configuration.
        cfgtitle = "PCIe Virtual Switch configurationt"
        mch_config[cfgtitle] = OrderedDict()

        # The first kind of rows are indexed by the index.
        indexes = [str(i) for i in range(6)]
        indexes = ["none"] + indexes
        indexes.append("Max. Link Speed")

        # Now, iterate over the rows of the table
        rows = forms[1].find_all("tr")

        for row in rows:
            # Let's skip all the initial rows with no data to be collected
            # Only the first th field is properly parsed, so this dirty trick
            # came up to cope with the issue:
            title = row.find("b")
            if title is None or (title is not None and title.text not in indexes):
                continue

            # Check whether the first select field is enabled in the row
            selects = row.find_all("select")

            # The row tagged with "none" has not select input fields
            if selects != []:
                # First, parse the Upstream option
                for s in selects:
                    for option in s.find_all("option"):
                        if "selected" in option.attrs:
                            mch_config[cfgtitle][s["name"]] = option["value"]

            # Now, iterate over the columns and check what radio buttons
            # are checked.

            for input in row.find_all("input"):
                if "checked" in input.attrs and "disabled" not in input.attrs:
                    name = input["name"]
                    mch_config[cfgtitle][name] = input["value"]

        # The previous set of parameters will be encapsulated so
        # a global dictionary can contain other set of parameters
        pciecfg = OrderedDict()
        pciecfg["PCIe parameter"] = mch_config

        return pciecfg

    def _flatten_cfg(self, cfgdict: OrderedDict) -> dict:
        """Flatten the configuration dictionary.

        The configuration is stored in a file using a human readable format.
        There are multiple level of nested dictionaries and IP addresses are
        stored using lists. This method flattens that structure and unrolls
        the variables that are lists.

        Args:
            cfgdict (OrderedDict): A subset of a dictionary as it is read from a
            configuration file.

        Returns:
            dict: A dictionary containing the flattened data from `cfgdict`
        """
        outdict = {}
        # The first layer can be seen as a collection of keys pointing to
        # sub-dictionaries.
        for okey in cfgdict:
            # At this level, we either have a set of pairs parameter-value,
            # or a list.
            for ikey in cfgdict[okey]:
                if not isinstance(cfgdict[okey][ikey], list):
                    outdict[ikey] = cfgdict[okey][ikey]
                else:
                    # Remove the last char (index)
                    nkey = ikey[:-1]
                    index = 0
                    for item in cfgdict[okey][ikey]:
                        outdict["{}{}".format(nkey, index)] = item
                        index += 1

        return outdict

    def _reboot(self, sleep: int = 60):
        """Reboot the MCH

        This internal method issues an HTTP POST request to the web page to perform
        a reboot of the device. The expected behaviour is not receiving any response
        to the HTTP POST, so an Exception should be catch when the device properly
        rebooted. Sometimes, the MCH takes too long to receive an IP address from
        the DHCP sever. This method will wait for about 2 minutes, then it will
        raise an Exception to indicate that the device couldn't be accessed. This
        situation might happen because a slow response of the DCHP server, or because
        something broke the MCH's configuration, so it fails to boot.

        Args:
            sleep: indicates how many seconds to wait after returning from the
            method. Write a 0 to avoid it.

        Returns:
            True if the MCH is accessible after issuing the reboot request and
            waiting for a timeout.
            False when the reboot request wasn't processed by the MCH.

        Raises:
            NoRouteToDevice when the MCH couldn't be accessed after the reboot.
        """
        self.logger.logger.info("Rebooting MCH. Sleeping for {} seconds.".format(sleep))

        data = {"save": "Reboot"}
        # The MCH doesn't return anything to the HTTP POST for the rebooting page
        # Use a timeout and check later that the MCH is up again
        try:
            rq.post(
                "http://{}/goform/reboot_web".format(self.ip_address),
                headers=self._http_headers,
                data=data,
                timeout=1,
            )
        except rq.ReadTimeout:
            # Wait until the MCH is up
            time.sleep(sleep)
            # Sometimes, the MCH takes longer to reboot
            for i in range(10):
                try:
                    self.open()
                    return True
                except Exception:
                    if i == 9:
                        msg = "The MCH didn't come back from the reboot."
                        self.logger.logger.error(msg)
                        raise NoRouteToDevice(msg)
                    else:
                        time.sleep(10)
                        continue

        # The expected flow is receiving the ReadTimeout exception when
        # the reboot was success
        return False

    def _check_arguments(self, data, target: str) -> tuple[bool, str]:
        """Check the content and type of the arguments for a method.

        Args:
            data (type): Description of parameter `data`.
            target (str): Description of parameter `target`.

        Returns:
            type: Description of returned object.
        """
        if not isinstance(data, eval(self._cfgkeys[target]["type"])):
            return (
                False,
                "The given data doesn't match the expectation. See the documentation.",
            )
        if self._cfgkeys[target]["cfgkey"] not in data.keys() and target is not None:
            return (
                False,
                "The given data doesn't match the expectation. See the documentation.",
            )
        return True, ""

    def device_info(self) -> tuple[bool, OrderedDict]:
        """Retrieve the main information about the device.

        The information is returned in a dictionary with 2 categories:
        *Board* and *Network*.
        This feature is supported by all the implemented communication
        interfaces, so the best one is chosen when multiple are allowed.

        An example of the returned dictionary::

            {
              'Board': {
                'fw_ver': 'V2.21.8',
                'fpga_ver': 'V1.14',
                'mcu_ver': '1.2',
                'serial_num': '113522-1426'},
              'Network': {'ip_address': '172.30.5.238',
                'mac_address': '00:40:42:22:05:92',
                'subnet_address': '255.255.252.0',
                'gateway_address': '172.30.7.254'}
            }

        Returns:
            A tuple containing the a success flag in the first position. In the \
            second position:
            - On success, a dictionary with the device information.
            - On failure, an empty dictionary.

        Raises:
            NoRouteToDevice: when the device is not reachable in the network, or the \
            device pointed by the given IP address is not an MCH.
        """
        if not self._link_established:
            self.open()

        response = rq.get(
            "http://{}/goform/GetInfo".format(self.ip_address),
            headers=self._http_headers,
        )

        if response.ok:
            resp_dict = self._parse_device_info(response)
            valid = True
        else:
            valid = False
            resp_dict = OrderedDict()

        return valid, resp_dict

    def update_fw(self, fw_path: str, part: str) -> bool:
        """Update the firmware of the device.

        This module doesn't support this feature. The main problem is that the
        process is divided in two stages. First, the binary is sent via post to
        the MCH. Then, the MCH analyses it and decides whether is a valid MCH
        firmware image or not. If so, it provides a checkbox and a new form so
        an user can click and proceed with the update. Automatizing this process
        produces an error in the MCH (*multiple access error*).
        I couldn't find a solution for that, which doesn't means it doesn't
        exists, because my experience with web programming is quite limited.
        """
        msg = "The firmware update is not supported by this module."
        self.logger.logger.info(msg)
        raise FeatureNotSupported(msg)

    def set_configuration(
        self, category: str, data: OrderedDict, apply: bool = True
    ) -> tuple[bool, str]:
        """Change the configuration of the device.

        This method focuses on the configuration parameters that are not
        defined within a configuration script. Specifying the entire set of
        parameters is not mandatory, and also, a particular category of
        settings can be modified without affecting the rest.

        This method supports the following configuration categories (taken
        from the webpage names):

        - Base Configuration [basecfg]
        - PCIe Virtual Switches [pcie]

        Pushing the backplane configuration file to the MCH using the Web
        interface doesn't seem doable. Some details could be found within
        this Jira ticket ICSHWI-7095.

        Args:
            category(str): the target set of parameters to be affected by the change. \
            The accepted values are previously listed.
            data(dic): dictionary containing the values to be modified.
            apply(bool): whether to reboot the MCH to make active the changes or not. \
            Disable rebooting when multiple changes are going to be performed in the device. \
            Reboot  the device after the last change.

        Returns:
            A tuple containing a `bool` indicating if the operation was successful,
            and an error message.

        Raises:
            NoRouteToDevice: when the device is not reachable after applying the \
            new configuration.
        """
        self.logger.logger.info("Setting {} configuration".format(category))

        if not self._link_established:
            self.open()

        validargs, _ = self._check_arguments(data, category)
        if not validargs:
            msg = "The given data doesn't match the expectation."
            self.logger.logger.error(msg)
            return (False, msg)

        # Check the input parameter
        if category == "basecfg":
            cfgword = self._cfgkeys[category]["webkey"]  # ["get_mch_cfg"]
            cfgkey = self._cfgkeys[category]["cfgkey"]  # "Base MCH parameter"
            # Data from a file contains extra keys to ease human readability that
            # are not needed by the MCH. The final dictionary has to be flattened.
            postdata = [self._flatten_cfg(data[cfgkey])]
        elif category == "pcie":
            # For the PCIe config, the dictionary has less hierarchy levels,
            # but it has to be split so each part of the configuration receives
            # only the expected parameters.
            cfgword = [
                "pcie_width_link_ctrl",
                "pcie_vs_cfg_refresh",
                "pcie_vs_cfg_backup_flash_save",
            ]
            cfgkey = "PCIe parameter"
            # if cfgkey not in data.keys():
            #     return (
            #         False,
            #         "The given data doesn't match the expectation. See the documentation.",
            #     )
            postdata = [
                {**data[cfgkey]["Link Width Configuration"]},
                {**data[cfgkey]["PCIe Virtual Switch configurationt"]},
                {},
            ]
        elif category == "backplane":
            msg = (
                "Changing the backplane configuration is not supported by this module."
            )
            self.logger.logger.error(msg)
            raise FeatureNotSupported(msg)
        else:
            cfgword = []

        if cfgword == []:
            msg = "The given category ({}) is not valid. See the documentation.".format(
                category
            )
            self.logger.logger.error(msg)
            return (False, msg)

        # And this key is needed at the end of the dictionary
        saveword = "Save"
        # An extra parameter has to be added to the post header
        headers = self._http_headers
        headers["Content-Type"] = "application/x-www-form-urlencoded"
        response = rq.Response()

        for word, pdata in zip(cfgword, postdata):
            pdata["save"] = saveword
            response = rq.post(
                "http://{}/goform/{}".format(self.ip_address, word),
                headers=headers,
                data=pdata,
            )
            if not response.ok:
                break
            saveword = "Apply" if saveword == "Save" else "Save"

        if response.ok:
            return_value = True, ""
            self.logger.logger.info(
                "Setting configuration of {} succeeded.".format(category)
            )
        else:
            msg = "Configuration update for '{}' failed. Check the logs.".format(
                category
            )
            self.logger.logger.error(msg)
            return_value = (False, msg)

        if apply:
            # This call might raise NoRouteToDevice, it should be handled upstream
            self._reboot()

        return return_value

    def get_configuration(self, category: str) -> tuple[bool, OrderedDict]:
        """Get the configuration of the device.

        This method returns a dictionary containing the configuration
        parameters of the device implementing this interface. If the device
        has several configuration categories, keys from the dictionary might
        contain other dictionaries.

        This method supports the following configuration categories (taken
        from the webpage names):

        - Base Configuration [basecfg]
        - PCIe Virtual Switch Configuration [pcie]
        - Backplane Configuration [backplane]

        Args:
            category(str): points to a subset of the configuration parameters of \
            the device. Use the values given between brackets from the previous \
            item list.

        Returns:
            tuple[bool, OrderedDict]: The first element indicates the success of \
            the operation, the second is an ordered dict containing the settings in \
            the same order as they are found in the web page. When *category* = \
            **backplane**, the dictionary contains only one key (*Backplane \
            config file*) and the whole configuration file as value for that \
            key.
        """
        self.logger.logger.info("Getting {} configuration".format(category))

        if not self._link_established:
            self.open()

        # Check the input parameter
        if category == "basecfg":
            cfgword = "change_mch_cfg"
        elif category == "pcie":
            cfgword = "pcie_width_link_ctrl"
        elif category == "backplane":
            # Request the generation of the script file to the MCH
            cfgword = "web_cfg_backup_show_menu"
        else:
            cfgword = ""

        mch_config: OrderedDict = OrderedDict()

        if cfgword == "":
            self.logger.logger.error(
                'Get configuration failed for unsupported category "{}".'.format(
                    category
                )
            )
            return False, mch_config

        response = rq.get(
            "http://{}/goform/{}".format(self.ip_address, cfgword),
            headers=self._http_headers,
        )

        if response.ok:
            if category != "backplane":
                # Let's parse the content
                parse_method = getattr(self, "_parse_{}".format(category))
                mch_config = parse_method(response)
            else:
                cfgword = "nat_mch_startup_cfg.txt"
                response = rq.get(
                    "http://{}/{}".format(self.ip_address, cfgword),
                    headers=self._http_headers,
                )
                mch_config["Backplane configuration"] = {
                    "file": (response.text if response.ok else "")
                }

        return True, mch_config

    def check_configuration(
        self, category: str, config: OrderedDict
    ) -> tuple[bool, str]:
        """Check the settings of the device.

        This method accesses the verification feature of the MCH's Web
        interface. This feature is only offered for the backplane configuration
        file. Use the NATMCH module for checking the rest of the configuration
        settings.

        This type of check only works when the backplane configuration file
        was set using the Web interface.

        Args:
            category (str): `backplane` is the only supported value.
            config (OrderedDict): the content of the backplane configuration file read in a \
            str. The key *Backplane config file* is expected, containing the whole \
            configuration file as a **str**.

        Returns:
            tuple[bool, str]: a boolean value indicating whether the
            checking was successful or not; and a text message when the
            checking was unsuccessful.
        """
        self.logger.logger.info("Checking {} configuration".format(category))

        if not self._link_established:
            self.open()

        validargs, _ = self._check_arguments(config, category)
        if not validargs:
            msg = "The given data doesn't match the expectation."
            self.logger.logger.error(msg)
            return (False, msg)

        # Check the input parameter
        if category == "backplane":
            cfgword = self._cfgkeys[category]["webkey"][0]
            cfgkey = self._cfgkeys[category]["cfgkey"]
        else:
            msg = "The given category ({}) is not valid. See the documentation.".format(
                category
            )
            return (False, msg)

        response = rq.post(
            "http://{}/goform/{}".format(self.ip_address, cfgword),
            headers=self._http_headers,
            files={"file": config[cfgkey]["file"]},
        )

        if response.ok:
            htmlsoup = BeautifulSoup(response.text, "lxml")
            validation_result = htmlsoup.find("div", {"class": "info"}).text

            if "PASSED" in validation_result:
                msg = "Configuration check for {} passed.\n{}".format(
                    category, validation_result
                )
                self.logger.logger.info(msg)
                return (True, validation_result)
        else:
            validation_result = "Bad request"

        return (False, validation_result)



# -*- coding: utf-8 -*-

"""
gendev_err.py
~~~~~~~~~~~~~

Custom exeception types for the GenDev Tools library.
"""

__all__ = [
    "ConnNotImplemented",
    "ConnTimeout",
    "NoRouteToDevice",
    "FeatureNotSupported",
]
__author__ = "Felipe Torres González"
__copyright__ = "Copyright 2021, ESS MCH Tools"
__credits__ = ["Felipe Torres González", "Ross Elliot", "Jeong Han Lee"]
__license__ = "GPL-3.0"
__version__ = "1.1"
__maintainer__ = "Ross Elliot"
__email__ = "ross.elliot@ess.eu"
__status__ = "Development"


class ConnNotImplemented(Exception):
    """Connection Not Implemented Exception.

    This exception is raised when an operation relies on a particular
    connection type that is not implemented for the given device.
    """

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "ConnNotImplemented, {0} ".format(self.message)
        else:
            return "ConnNotImplemented has been raised"


class ConnTimeout(Exception):
    """Connection timeout Exception."""

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "ConnTimeout, {0} ".format(self.message)
        else:
            return "ConnTimeout has been raised"


class NoRouteToDevice(Exception):
    """No route to device Exception."""

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "NoRouteToDevice, {0} ".format(self.message)
        else:
            return "NoRouteToDevice has been raised"


class ConnectionRefused(Exception):
    """Connection refused Exception."""

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "ConnectionRefused, {0} ".format(self.message)
        else:
            return "ConnectionRefused has been raised"


class FeatureNotSupported(Exception):
    """Feature not supported exception.

    This exception is raised when a feature is not possible with the given
    valid communication interfaces to the MCH.
    """

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "ConnTimeout, {0} ".format(self.message)
        else:
            return "ConnTimeout has been raised"


class WebChanged(Exception):
    """The content of the Web page has changed.

    This exception is raised when the parser fails analysing the content of the
    web interface of the device.
    """

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return "WebChanged, {0} ".format(self.message)
        else:
            return "WebChanged has been raised"


class DHCPEnableFailed(Exception):
    """DHCP mode enable failed exception.

    This exception is raised when the enabling of DHCP mode
    has failed.
    """

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return " DHCPEnableFailed, {0} ".format(self.message)
        else:
            return "DHCPEnableFailed"
